/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <ctime>
#include <vector>
#include <utility>

#include "idlc/cpp_gen/cpp_gen_base.h"

namespace {
#include "idlc/cpp_gen/cpp_gen_base_cb.h"
}

namespace tidl {

CppGeneratorBase::CppGeneratorBase(std::shared_ptr<Document> doc)
    : Generator(doc) {
  type_map_ = {
      {"char", "char"}, {"int", "int"}, {"short", "short"},
      {"long", "long long"}, {"string", "std::string"}, {"bool", "bool"},
      {"list", "std::list"}, {"float", "float"}, {"double", "double"},
      {"bundle", "Bundle"}, {"void", "void"}, {"array", "std::vector"}
  };

  parcel_type_map_ = {
    {"char", "byte"},
    {"int", "int32"},
    {"short", "int16"},
    {"long", "int64"},
    {"string", "string"},
    {"bool", "bool"},
    {"float", "float"},
    {"double", "double"},
    {"bundle", "bundle"},
  };

  type_init_map_ = {
    {"char", "0"},
    {"int", "0"},
    {"short", "0"},
    {"long", "0"},
    {"bool", "false"},
    {"float", "0.0f"},
    {"double", "0.0"},
  };
}

void CppGeneratorBase::GenStructuresForHeader(std::ofstream& stream) {
  stream << CB_BUNDLE;
  for (auto& i : GetDocument().GetBlocks()) {
    if (i->GetType() != Block::TYPE_STRUCTURE)
      continue;
    Structure& st = static_cast<Structure&>(*i);
    GenStructureForHeader(stream, st);
    stream << std::endl;
  }
}

void CppGeneratorBase::GenStructureForHeader(std::ofstream& stream,
                                             const Structure& st) {
  const char ctor[] = "  $$();\n" \
                      "  $$($$);\n";
  const char variable[] = "$$\n";

  stream << "class " << st.GetID() << " final ";

  GenBrace(stream, 0, [&]() {
    stream << " public:" << NLine(1);
    GenTemplate(ctor, stream,
      [&]()->std::string {
        return st.GetID();
      },
      [&]()->std::string {
        return st.GetID();
      },
      [&]()->std::string {
        std::string str;
        int n = 1;
        for (auto& i : st.GetElements().GetElms()) {
          if (n != 1)
            str += ", ";
          str += ConvertTypeToString(i->GetType()) + " " + i->GetID();
          n++;
        }
        return str;
      });

    stream << NLine(1);
    for (auto& i : st.GetElements().GetElms()) {
      GenSetter(stream, *i);
      GenGetter(stream, *i);
      stream << NLine(1);
    }

    stream << " private:";
    GenTemplate(variable, stream,
      [&]()->std::string {
        std::string str;
        for (auto& i : st.GetElements().GetElms()) {
          str += NLine(1) + Tab(1)
              + ConvertTypeToString(i->GetType()) + " "
              + i->GetID() + "_";
          if (type_init_map_.find(i->GetType().ToString())
              == type_init_map_.end()) {
            str += ";";
          } else {
            str += " = " + type_init_map_[i->GetType().ToString()] + ";";
          }
        }
        str += NLine(1);
        return str;
      });
  }, false, false);
  stream << ";" << NLine(1);
}

void CppGeneratorBase::GenSetter(std::ofstream& stream, const Element& ele) {
  const char setter[] =
      "void Set$$($$ $$) {\n" \
      "  $$_ = $$;\n" \
      "}\n";

  GenTemplate(AddIndent(TAB_SIZE, setter, true), stream,
    [&]()->std::string {
      return ele.GetID();
    },
    [&]()->std::string {
      return ConvertTypeToString(ele.GetType());
    },
    [&]()->std::string {
      return ele.GetID();
    },
    [&]()->std::string {
      return ele.GetID();
    },
    [&]()->std::string {
      if (ele.GetType().IsUserDefinedType() ||
        ele.GetType().GetMetaType() != nullptr ||
        ele.GetType().ToString() == "string" ||
        ele.GetType().ToString() == "bundle") {
        return "std::move(" + ele.GetID() + ")";
      }

      return ele.GetID();
    });
  stream << NLine(1);
}

void CppGeneratorBase::GenGetter(std::ofstream& stream, const Element& ele) {
  const char getter[] =
      "$$ Get$$() const {\n" \
      "  return $$_;\n" \
      "}\n";

  GenTemplate(AddIndent(TAB_SIZE, getter, true), stream,
    [&]()->std::string {
      if (ele.GetType().IsUserDefinedType() ||
        ele.GetType().GetMetaType() != nullptr ||
        ele.GetType().ToString() == "string" ||
        ele.GetType().ToString() == "bundle") {
        return "const " + ConvertTypeToString(ele.GetType()) + "&";
      }

      return ConvertTypeToString(ele.GetType());
    },
    [&]()->std::string {
      return ele.GetID();
    },
    [&]()->std::string {
      return ele.GetID();
    });
}

void CppGeneratorBase::GenStructuresForBody(std::ofstream& stream) {
  for (auto& i : GetDocument().GetBlocks()) {
    if (i->GetType() != Block::TYPE_STRUCTURE)
      continue;
    Structure& st = static_cast<Structure&>(*i);
    GenStructureForBody(stream, st);
    stream << std::endl;
  }
}

void CppGeneratorBase::GenStructureForBody(std::ofstream& stream,
    const Structure& st) {
  std::vector<std::pair<std::string, std::string>> v;
  const char ctor[] = "##::##() {}\n\n" \
                      "##::##($$)\n" \
                      "    : $$ {}";

  for (auto& i : st.GetElements().GetElms()) {
    std::pair<std::string, std::string> p;

    p.first = ConvertTypeToString(i->GetType());
    p.second = i->GetID();
    v.push_back(p);
  }

  GenTemplate(ReplaceAll(ctor, "##", st.GetID()), stream,
    [&]()->std::string {
      std::string str;
      for (auto& i : v) {
        str += i.first + " " + i.second;

        if (i != v.back())
          str += ", ";
      }
      return str;
    },
    [&]()->std::string {
      std::string str;
      for (auto& i : v) {
        str += i.second + "_(std::move(" + i.second + "))";

        if (i != v.back())
          str += ", ";
      }
      return str;
    });
  stream <<  NLine(2);
}

void CppGeneratorBase::GenSerializer(std::ofstream& stream) {
  for (auto& i : GetDocument().GetBlocks()) {
    if (i->GetType() != Block::TYPE_STRUCTURE)
      continue;
    Structure& st = static_cast<Structure&>(*i);
    GenSerializer(stream, st);
    stream << NLine(1);
  }
}

void CppGeneratorBase::GenPrototype(std::ofstream& stream) {
  for (auto& i : GetDocument().GetBlocks()) {
    if (i->GetType() != Block::TYPE_STRUCTURE)
      continue;
    Structure& st = static_cast<Structure&>(*i);
    GenSerializer(stream, st, true);
    GenDeSerializer(stream, st, true);
  }
  GenListSerializer(stream, true);
  stream << NLine(1);
}

void CppGeneratorBase::GenSerializer(std::ofstream& stream, const Structure& st,
                                     bool proto) {
  const char parcel_str[] = "rpc_port_parcel_h";

  stream << parcel_str << " operator << ("
         << parcel_str << " h, const " << st.GetID() << "& param)";
  if (proto) {
    stream << ";" << NLine(1);
    return;
  }

  stream << " ";
  GenBrace(stream, 0, [&]() {
    for (auto& i : st.GetElements().GetElms()) {
      stream << AddIndent(TAB_SIZE,
          ConvertTypeToSerializer(i->GetType(),
              "param.Get" + i->GetID() + "()", "h"));
    }
    stream << Tab(1) << "return h;" << NLine(1);
  }, false);
}

void CppGeneratorBase::GenDeSerializer(std::ofstream& stream) {
  for (auto& i : GetDocument().GetBlocks()) {
    if (i->GetType() != Block::TYPE_STRUCTURE)
      continue;
    Structure& st = static_cast<Structure&>(*i);
    GenDeSerializer(stream, st);
    stream << std::endl;
  }
}

void CppGeneratorBase::GenDeSerializer(std::ofstream& stream,
                                       const Structure& st, bool proto) {
  const char parcel_str[] = "rpc_port_parcel_h";

  stream << parcel_str << " operator >> ("
         << parcel_str << " h, " << st.GetID() << "& param)";
  if (proto) {
    stream << ";" << NLine(1);
    return;
  }

  stream << " ";
  GenBrace(stream, 0, [&]() {
    for (auto& i : st.GetElements().GetElms()) {
      stream << AddIndent(TAB_SIZE,
          ConvertTypeToDeserializer(i->GetType(), i->GetID(), "h"));
      stream << Tab(1) << "param.Set" << i->GetID() << "(std::move("
             << i->GetID() << "));" << NLine(2);
    }
    stream << Tab(1) << "return h;" << NLine(1);
  }, false);
}

std::string CppGeneratorBase::ConvertTypeToString(const BaseType& type) {
  if (type.IsUserDefinedType()) {
    if (IsDelegateType(type)) {
      return "std::unique_ptr<" + type.ToString() + ">";
    }
    return type.ToString();
  }

  if (type.GetMetaType() != nullptr)
    return type_map_[type.ToString()] + "<" +
        ConvertTypeToString(*(type.GetMetaType())) + ">";

  return type_map_[type.ToString()];
}

std::string CppGeneratorBase::Tab(int cnt) {
  std::string t(cnt * 2, ' ');

  return t;
}

std::string CppGeneratorBase::NLine(int cnt) {
  std::string t(cnt, '\n');

  return t;
}

void CppGeneratorBase::AddSerializerList(const BaseType& type) {
  if (type.GetMetaType() != nullptr) {
    serializer_list_[ConvertTypeToString(type)] = &type;
    AddSerializerList(*type.GetMetaType());
  }
}

void CppGeneratorBase::GenListSerializer(std::ofstream& stream,
                                         const BaseType& type, bool proto) {
  stream << "rpc_port_parcel_h operator << (rpc_port_parcel_h h, const "
         << ConvertTypeToString(type) << "& c)";

  if (proto) {
    stream << ";" << NLine(1);
    stream << "rpc_port_parcel_h operator >> (rpc_port_parcel_h h, "
           << ConvertTypeToString(type) << "& c);" << NLine(1);
    return;
  }

  stream << " ";
  GenBrace(stream, 0, [&]() {
    stream << Tab(1)
           << "rpc_port_parcel_write_array_count(h, c.size());"
           << NLine(1);
    stream << Tab(1) << "for (auto& i : c) ";
    GenBrace(stream, TAB_SIZE, [&]() {
      auto& mt = *type.GetMetaType();
      stream << AddIndent(TAB_SIZE * 2, ConvertTypeToSerializer(mt, "i", "h"));
    }, false);
    stream << Tab(1) << "return h;" << NLine(1);
  }, false);
  stream << NLine(1);

  stream << "rpc_port_parcel_h operator >> (rpc_port_parcel_h h, "
         << ConvertTypeToString(type) << "& c) ";
  GenBrace(stream, 0, [&]() {
    stream << Tab(1) << "int l = 0;" << NLine(1);
    stream << Tab(1)
           << "rpc_port_parcel_read_array_count(h, &l);" << NLine(1);
    stream << Tab(1) << "for (int i = 0; i < l; i++) ";
    GenBrace(stream, TAB_SIZE, [&]() {
      auto& mt = *type.GetMetaType();
      stream << AddIndent(TAB_SIZE * 2,
          ConvertTypeToDeserializer(mt, "v", "h", true));
      stream << Tab(2) << "c.push_back(std::move(v));" << NLine(1);
    }, false);
    stream << Tab(1) << "return h;" << NLine(1);
  }, false);
}

void CppGeneratorBase::GenListSerializer(std::ofstream& stream, bool proto) {
  serializer_list_.clear();
  for (auto& i : GetDocument().GetBlocks()) {
    if (i->GetType() == Block::TYPE_STRUCTURE) {
      const Structure& st = static_cast<const Structure&>(*i);
      for (auto& j : st.GetElements().GetElms()) {
        auto& t = j->GetType();
        AddSerializerList(t);
      }
    } else if (i->GetType() == Block::TYPE_INTERFACE) {
      const Interface& iface = static_cast<const Interface&>(*i);
      for (auto& j : iface.GetDeclarations().GetDecls()) {
        auto& t = j->GetType();
        AddSerializerList(t);
        for (auto& k : j->GetParameters().GetParams()) {
          auto& t1 = k->GetParameterType().GetBaseType();
          AddSerializerList(t1);
        }
      }
    }
  }

  for (auto& p : serializer_list_) {
    const BaseType* t = p.second;
    GenListSerializer(stream, *t, proto);
  }
}

void CppGeneratorBase::GenMethodId(std::ofstream& stream,
                                   const Interface& iface) {
  stream << Tab(1) << "enum class MethodId : int ";
  GenBrace(stream, TAB_SIZE, [&]() {
    int cnt = 2;
    stream << Tab(2) << "__Result = 0," << NLine(1);
    stream << Tab(2) << "__Callback = 1," << NLine(1);
    for (auto& i : iface.GetDeclarations().GetDecls()) {
      if (i->GetMethodType() == Declaration::MethodType::DELEGATE)
        continue;
      stream << Tab(2)
             << i->GetID() << " = " << cnt++ << "," << NLine(1);
    }
  }, false, false);
  stream << ";" << NLine(2);
}

void CppGeneratorBase::GenDelegateId(std::ofstream& stream,
                                     const Interface& iface) {
  stream << Tab(1) << "enum class DelegateId : int ";
  GenBrace(stream, TAB_SIZE, [&]() {
    int cnt = 1;
    for (auto& i : iface.GetDeclarations().GetDecls()) {
        if (i->GetMethodType() != Declaration::MethodType::DELEGATE)
          continue;
        stream << Tab(2)
               << i->GetID() << " = " << cnt++ << "," << NLine(1);
    }
  }, false, false);
  stream << ";" << NLine(2);
}

void CppGeneratorBase::GenParameters(std::ofstream& stream,
                                     const Parameters& ps) {
  stream << GetParameters(ps);
}

std::string CppGeneratorBase::GetParameters(const Parameters& ps) {
  bool first = true;
  std::string ret;
  for (auto& i : ps.GetParams()) {
    if (!first) {
      ret += ", ";
    }

    std::string ref;
    auto dir = i->GetParameterType().GetDirection();
    if (dir == ParameterType::Direction::OUT ||
        dir == ParameterType::Direction::REF) {
      ref = "&";
    }

    ret += ConvertTypeToString(i->GetParameterType().GetBaseType())
        + ref + " " + i->GetID();
    first = false;
  }

  return ret;
}

void CppGeneratorBase::GenDeclaration(std::ofstream& stream,
                                     const Interface& iface,
                                     const Declaration& decl) {
  stream << ConvertTypeToString(decl.GetType()) << " " << iface.GetID() << "::"
         << decl.GetID() << "(";
  GenParameters(stream, decl.GetParameters());
  stream << ") ";
}

std::string CppGeneratorBase::ConvertTypeToSerializer(
    const BaseType& type, std::string id, std::string parcel) {
  std::string ret;

  if (type.ToString() == "string") {
    ret += "rpc_port_parcel_write_"
        + parcel_type_map_[type.ToString()]
        + "(" + parcel + ", " + id + ".c_str());\n";
  } else if (type.ToString() == "bundle") {
    ret += "rpc_port_parcel_write_bundle(" + parcel + ", "
        + id + ".GetHandle());\n";
  } else if (type.GetMetaType() || type.IsUserDefinedType()) {
    if (type.IsUserDefinedType() && IsDelegateType(type))
      ret += parcel + " << *" + id + ";\n";
    else
      ret += parcel + " << " + id + ";\n";
  } else {
    ret += "rpc_port_parcel_write_"
        + parcel_type_map_[type.ToString()]
        + "(" + parcel + ", " + id + ");\n";
  }

  return ret;
}

std::string CppGeneratorBase::ConvertTypeToDeserializer(
    const BaseType& type, std::string id, std::string parcel,
    bool make_new_type) {
  std::string ret;

  if (type.ToString() == "string") {
    ret += "char* " + id + "_raw  = nullptr;\n";
    ret += "rpc_port_parcel_read_" + parcel_type_map_[type.ToString()]
        + "(" + parcel + ", &" + id + "_raw);\n";
    if (make_new_type) {
      ret += "std::string " + id + "(" + id + "_raw);\n";
    } else {
      ret += id + " = " + id + "_raw;\n";
    }
    ret += "free(" + id + "_raw);\n";
  } else if (type.ToString() == "bundle") {
    ret += "bundle* " + id + "_raw = nullptr;\n";
    ret += "rpc_port_parcel_read_" + parcel_type_map_[type.ToString()]
        + "(" + parcel + ", &" + id + "_raw);\n";
    if (make_new_type) {
      ret += "Bundle " + id + "(" + id + "_raw);\n";
    } else {
      ret +=  id + " = " + id + "_raw;\n"
          + "bundle_free(" + id + "_raw);\n";
    }
  } else if (type.GetMetaType() != nullptr || type.IsUserDefinedType()) {
    std::string n;

    if (type.GetMetaType() != nullptr || IsDelegateType(type))
      n = ConvertTypeToString(type);
    else
      n = type.ToString();

    if (make_new_type) {
      ret += n + " ";
      if (IsDelegateType(type)) {
        ret += id + "(new " + type.ToString()
            + "(port, std::weak_ptr<ServiceBase>(b)));\n";
      } else {
        ret += id + ";\n";
      }
    }
    if (IsDelegateType(type))
      ret += parcel + " >> *" + id + ";\n";
    else
      ret += parcel + " >> " + id + ";\n";
  } else {
    if (make_new_type)
      ret += ConvertTypeToString(type) + " " + id + ";\n";
    ret += "rpc_port_parcel_read_" + parcel_type_map_[type.ToString()]
        + "(" + parcel + ", &" + id + ");\n";
  }

  return ret;
}

void CppGeneratorBase::GenBodyCallbacks(std::ofstream& stream,
                                   const Interface& iface, bool is_proxy) {
  stream << ReplaceAll(CB_CALLBACK_BASE, "##", iface.GetID());

  for (auto& i : iface.GetDeclarations().GetDecls()) {
    if (i->GetMethodType() != Declaration::MethodType::DELEGATE)
          continue;
    GenBodyCallback(stream, iface, *i, is_proxy);
  }
}

void CppGeneratorBase::GenBodyCallback(std::ofstream& stream,
    const Interface& iface, const Declaration& decl, bool is_proxy) {
  if (!is_proxy) {
    GenTemplate(CB_CALLBACK_INVOKE_METHOD, stream,
      [&]()->std::string {
        return iface.GetID();
      },
      [&]()->std::string {
        return decl.GetID();
      },
      [&]()->std::string {
        return GetParameters(decl.GetParameters());
      },
      [&]()->std::string {
        std::string m;
        for (auto& i : decl.GetParameters().GetParams()) {
          auto& pt = i->GetParameterType();
          m += AddIndent(TAB_SIZE,
              ConvertTypeToSerializer(pt.GetBaseType(), i->GetID(), "p"));
        }
        return m;
      });
  } else {
    GenTemplate(CB_CALLBACK_ON_RECEIVED_EVENT_METHOD, stream,
      [&]()->std::string {
        return iface.GetID();
      },
      [&]()->std::string {
        return decl.GetID();
      },
      [&]()->std::string {
        int cnt = 1;
        std::string ret;
        for (auto& i : decl.GetParameters().GetParams()) {
          std::string v = "param" + std::to_string(cnt);
          std::string c = ConvertTypeToDeserializer(
              i->GetParameterType().GetBaseType(), v, "parcel");
          ret += AddIndent(TAB_SIZE, c) + NLine(1);
          cnt++;
        }

        cnt = 1;
        ret += Tab(1) + "OnReceived(";
        for (int i = 0; i < decl.GetParameters().GetParams().size(); i++) {
          if (cnt != 1) {
            ret += ", ";
          }
          ret += "std::move(param" + std::to_string(cnt) + ")";
          cnt++;
        }
        ret += ");";

        return ret;
      });
  }
}

void CppGeneratorBase::GenHeaderCallbacks(std::ofstream& stream,
                                          const Interface& iface,
                                          bool is_proxy) {
  stream << CB_CALLBACK_BASE_HEADER_FRONT;
  if (is_proxy) {
    stream << Tab(1)
           << "  virtual void OnReceivedEvent(rpc_port_parcel_h port) = 0;"
           << NLine(1);
  }
  stream << CB_CALLBACK_BASE_HEADER_BACK;

  for (auto& i : iface.GetDeclarations().GetDecls()) {
    if (i->GetMethodType() != Declaration::MethodType::DELEGATE)
          continue;
    GenHeaderCallback(stream, *i, is_proxy);
  }
}

void CppGeneratorBase::GenHeaderCallback(std::ofstream& stream,
                                         const Declaration& decl,
                                         bool is_proxy) {
  stream << GenTemplateString(CB_CALLBACK_CLASS,
      [&]()->std::string {
        return decl.GetID();
      },
      [&]()->std::string {
        return ReplaceAll(
            is_proxy ? CB_CALLBACK_CTOR_PROXY : CB_CALLBACK_CTOR_STUB,
            "##", decl.GetID());
      },
      [&]()->std::string {
        std::string ret;
        if (is_proxy) {
          ret = Tab(2) + "virtual void OnReceived("
              + GetParameters(decl.GetParameters())
              + ") {}" + NLine(1);
        } else {
          ret = Tab(2) + "void Invoke("
              + GetParameters(decl.GetParameters())
              + ");" + NLine(1);
        }

        return ret;
      },
      [&]()->std::string {
        return is_proxy ? CB_CALLBACK_PRIVATE_PROXY : CB_CALLBACK_PRIVATE_STUB;
      });
}

void CppGeneratorBase::GenCopyright(std::ofstream& stream) {
  GenTemplate(CB_COPYRIGHT, stream,
    [&]()->std::string {
      return FULLVER;
    },
    [&]()->std::string {
      std::time_t t = std::time(NULL);
      std::tm* local_time = std::localtime(&t);
      return std::to_string(local_time->tm_year + 1900);
    });
}

void CppGeneratorBase::GenLogTag(std::ofstream& stream, std::string id) {
  GenTemplate(CB_LOG_TAG, stream,
      [&]()->std::string {
        return id;
      });
}

void CppGeneratorBase::GenLogDefinition(std::ofstream& stream) {
  stream << CB_LOG_DEF;
}

}  // namespace tidl
