/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "idlc/cpp_gen/cpp_stub_body_gen.h"

namespace {
#include "idlc/cpp_gen/cpp_stub_body_gen_cb.h"
}

namespace tidl {

CppStubBodyGen::CppStubBodyGen(std::shared_ptr<Document> doc)
    : CppGeneratorBase(doc) {}

void CppStubBodyGen::OnInitGen(std::ofstream& stream) {
  std::string key(".cc");
  std::string header_file = FileName;

  std::size_t found = header_file.rfind(key);
  if (found != std::string::npos)
    header_file.replace(found, key.length(), ".h");

  GenCopyright(stream);
  stream << NLine(1);
  stream << "#include <stdlib.h>" << NLine(1)
         << "#include <assert.h>" << NLine(1)
         << "#include <dlog.h>" << NLine(1)
         <<  NLine(1)
         << "#include \"" << header_file << "\"" << NLine(2);
  GenLogTag(stream, "RPC_PORT_STUB");
  GenLogDefinition(stream);
  stream << NLine(1);
  GenNamespace(stream);
}

void CppStubBodyGen::OnFiniGen(std::ofstream& stream) {}

void CppStubBodyGen::GenNamespace(std::ofstream& stream) {
  stream << "namespace rpc_port ";
  GenBrace(stream, 0, [&]() {
    stream << "namespace " << GetFileNamespace() << " ";
    GenBrace(stream, 0, [&]() {
      stream <<  NLine(1);
      GenStructuresForBody(stream);
      stream << "namespace stub ";
      GenBrace(stream, 0, [&]() {
        GenPrototype(stream);
        GenSerializer(stream);
        GenDeSerializer(stream);
        GenListSerializer(stream);
        GenInterfaces(stream);
      }, false, false);
      stream << "  // namespace stub" + NLine(1);
    }, false, false);
    stream << "  // namespace " + GetFileNamespace() + NLine(1);
  }, false, false);
  stream << "  // namespace rpc_port" + NLine(1);
}

void CppStubBodyGen::GenInterfaces(std::ofstream& stream) {
  for (auto& i : GetDocument().GetBlocks()) {
    if (i->GetType() != Block::TYPE_INTERFACE)
      continue;
    Interface& iface = static_cast<Interface&>(*i);
    GenInterface(stream, iface);
  }
}

void CppStubBodyGen::GenInterface(std::ofstream& stream,
                                    const Interface& iface) {
  GenServiceBase(stream, iface);
  GenBodyCallbacks(stream, iface, false);
  GenDefaultMethods(stream, iface);
  GenReceivedEvent(stream, iface);
}

void CppStubBodyGen::GenServiceBase(std::ofstream& stream,
                                    const Interface& iface) {
  GenTemplate(CB_CTOR_SERVICE_BASE, stream,
    [&]()->std::string {
      return iface.GetID();
    });
  stream << NLine(1);
}

void CppStubBodyGen::GenConstructor(std::ofstream& stream,
                                    const Interface& iface) {
  stream << ReplaceAll(CB_CTOR_FRONT, "##", iface.GetID());

  for (auto& i : iface.GetAttributes().GetAttrs()) {
    if (i->GetKey() == "privilege") {
      stream << Tab(1) << "rpc_port_stub_add_privilege(stub_, \""
             << i->GetValue() << "\");" << NLine(1);
    } else if (i->GetKey() == "trusted" && i->GetValue() == "true") {
      stream << Tab(1) << "rpc_port_stub_set_trusted(stub_, "
             << i->GetValue() << ");" << NLine(1);
    }
  }

  stream << "}" << NLine(1);
}

void CppStubBodyGen::GenDefaultMethods(std::ofstream& stream,
                                       const Interface& iface) {
  GenConstructor(stream, iface);
  stream << ReplaceAll(CB_DEFAULT_METHODS, "##", iface.GetID());
}

void CppStubBodyGen::GenReceivedEvent(std::ofstream& stream,
                                      const Interface& iface) {
  GenTemplate(CB_ON_RECEIVED_CB_FRONT, stream,
    [&]()->std::string {
      return iface.GetID();
    },
    [&]()->std::string {
      return iface.GetID();
    });
  for (auto& i : iface.GetDeclarations().GetDecls()) {
    if (i->GetMethodType() == Declaration::MethodType::DELEGATE)
      continue;
    stream << Tab(2) << "case static_cast<int>(MethodId::"
           << i->GetID() << "): ";
    GenBrace(stream, TAB_SIZE * 2, [&]() {
      GenInvocation(stream, *i);
      stream << Tab(3) << "break;" << NLine(1);
    }, false);
    stream << NLine(1);
  }
  stream << CB_ON_RECEIVED_CB_BACK << NLine(1);
}

void CppStubBodyGen::GenInvocation(std::ofstream& stream,
    const Declaration& decl) {
  int cnt = 1;

  // Deserialize
  for (auto& i : decl.GetParameters().GetParams()) {
    if (i->GetParameterType().GetDirection() == ParameterType::Direction::OUT) {
      cnt++;
      continue;
    }

    std::string v = "param" + std::to_string(cnt);
    std::string c = ConvertTypeToDeserializer(
        i->GetParameterType().GetBaseType(), v, "p");
    stream << AddIndent(TAB_SIZE * 3, c);
    cnt++;
  }

  // Invoke
  cnt = 1;
  std::string m;
  std::string d;
  bool hasRet = false;

  if (decl.GetType().ToString() != "void") {
    m += "auto retVal = ";
    hasRet = true;
  }

  m += "b->" + decl.GetID() + "(";
  for (auto& i : decl.GetParameters().GetParams()) {
    if (cnt != 1) {
      m += ", ";
    }

    std::string v = "param" + std::to_string(cnt);
    auto& pt = i->GetParameterType();
    if (pt.GetDirection() == ParameterType::Direction::OUT) {
      d += ConvertTypeToString(pt.GetBaseType()) + " " + v + ";\n";
    }
    if (IsDelegateType(pt.GetBaseType())) {
      m += "std::move(";
      m += v;
      m += ")";
    } else {
      m += v;
    }
    cnt++;
  }

  m += ");\n";
  stream << AddIndent(TAB_SIZE * 3, d);
  stream << AddIndent(TAB_SIZE * 3, m);

  // Serialize
  if (decl.GetMethodType() == Declaration::MethodType::ASYNC)
    return;

  cnt = 0;
  m = "rpc_port_parcel_write_int32(" \
      "result, static_cast<int>(MethodId::__Result));\n";
  for (auto& i : decl.GetParameters().GetParams()) {
    auto& pt = i->GetParameterType();
    cnt++;
    if (pt.GetDirection() == ParameterType::Direction::IN)
      continue;
    m += ConvertTypeToSerializer(pt.GetBaseType(),
                                "param" + std::to_string(cnt), "result");
  }

  if (hasRet) {
    m += ConvertTypeToSerializer(decl.GetType(), "retVal", "result");
  }

  m += "rpc_port_parcel_send(result, port);\n";
  stream << AddIndent(TAB_SIZE * 3, m);
}

}  // namespace tidl
