/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <glib.h>

#include <iostream>
#include <cstring>
#include <memory>

#include "idlc/parser.h"
#include "idlc/cs_gen/cs_proxy_gen.h"
#include "idlc/cs_gen/cs_stub_gen.h"
#include "idlc/cs_gen/cs_lib_gen.h"
#include "idlc/c_gen/c_proxy_header_gen.h"
#include "idlc/c_gen/c_proxy_body_gen.h"
#include "idlc/c_gen/c_stub_header_gen.h"
#include "idlc/c_gen/c_stub_body_gen.h"
#include "idlc/cpp_gen/cpp_proxy_header_gen.h"
#include "idlc/cpp_gen/cpp_proxy_body_gen.h"
#include "idlc/cpp_gen/cpp_stub_header_gen.h"
#include "idlc/cpp_gen/cpp_stub_body_gen.h"

namespace {

class Options {
 public:
  ~Options() = default;

  static std::unique_ptr<Options> Parse(int argc, char** argv);
  bool IsProxy() const { return isProxy_; }
  std::string GetLanguage() const { return language_; }
  std::string GetInput() const { return input_; }
  std::string GetOutput() const { return output_; }
  bool HasNamespace() const { return hasNamespace_; }
  bool HasRpcPortLib() const { return hasRpcPortLib_; }

 private:
  enum Cmd {
    CMD_PROXY,
    CMD_STUB,
    CMD_VERSION,
    CMD_MAX
  };

  enum Opt {
    OPT_LANGUAGE,
    OPT_INPUT,
    OPT_OUTPUT,
    OPT_NAMESPACE,
    OPT_RPCLIB,
    OPT_MAX
  };

  void PrintUsage();
  void PrintVersion();

 private:
  bool isProxy_;
  std::string language_;
  std::string input_;
  std::string output_;
  std::string help_;
  bool hasNamespace_ = false;
  bool hasRpcPortLib_ = false;
};

void Options::PrintUsage() {
  std::cerr << help_ << std::endl;
}

void Options::PrintVersion() {
  std::cerr << "tidlc " << FULLVER << std::endl;
}

std::unique_ptr<Options> Options::Parse(int argc, char** argv) {
  gpointer cmd[CMD_MAX] = { 0, };
  static GOptionEntry cmdEntries[] = {
    { "proxy", 'p', 0, G_OPTION_ARG_NONE, &cmd[CMD_PROXY],
       "Generate proxy code", NULL },
    { "stub", 's', 0, G_OPTION_ARG_NONE, &cmd[CMD_STUB],
      "Generate stub code", NULL },
    { "version", 'v', 0, G_OPTION_ARG_NONE, &cmd[CMD_VERSION],
      "Show version information", NULL },
    { NULL }
  };
  gpointer opt[OPT_MAX] = { 0, };
  static GOptionEntry optEntries[] = {
    { "language", 'l', 0, G_OPTION_ARG_STRING, &opt[OPT_LANGUAGE],
      "Select generating language (C, C++, C#).", "LANGUAGE" },
    { "input", 'i', 0, G_OPTION_ARG_STRING, &opt[OPT_INPUT],
      "A tidl interface file.", "INPUT" },
    { "output", 'o', 0, G_OPTION_ARG_STRING, &opt[OPT_OUTPUT],
      "The generated interface file.", "OUTPUT" },
    { "namespace", 'n', 0, G_OPTION_ARG_NONE, &opt[OPT_NAMESPACE],
      "Add the prefix in the funtion name as output file name " \
      "(C language only).",
      NULL },
    { "rpclib", 'r', 0, G_OPTION_ARG_NONE, &opt[OPT_RPCLIB],
      "Generate C# library for rpc-port " \
      "(C# language only).",
      NULL },
    { NULL }
  };

  GOptionContext* context = g_option_context_new(NULL);
  if (!context) {
    std::cerr << "Failed to create context" << std::endl;
    return std::unique_ptr<Options>(nullptr);
  }
  g_option_context_add_main_entries(context, cmdEntries, NULL);

  GOptionGroup* optGroup = g_option_group_new("option", "Additional Options:",
      "Additional options", NULL, NULL);
  if (!optGroup) {
    std::cerr << "Failed to create option group" << std::endl;
    g_option_context_free(context);
    return std::unique_ptr<Options>(nullptr);
  }
  g_option_group_add_entries(optGroup, optEntries);
  g_option_context_add_group(context, optGroup);

  GError* err = NULL;
  if (!g_option_context_parse(context, &argc, &argv, &err)) {
    std::cerr << argv[0] << ": " << err->message << std::endl;
    g_option_context_free(context);
    g_clear_error(&err);
    return std::unique_ptr<Options>(nullptr);
  }

  gchar* help = g_option_context_get_help(context, TRUE, NULL);

  std::unique_ptr<Options> options(new Options());
  options->help_ = help;

  g_free(help);
  g_option_context_free(context);

  if (cmd[CMD_VERSION]) {
    options->PrintVersion();
    return std::unique_ptr<Options>(nullptr);
  }

  if ((!cmd[CMD_PROXY] && !cmd[CMD_STUB]) ||
       !opt[OPT_LANGUAGE] || !opt[OPT_INPUT] ||
       !opt[OPT_OUTPUT]) {
    options->PrintUsage();
    return std::unique_ptr<Options>(nullptr);
  }

  options->isProxy_ = cmd[CMD_PROXY] ? true : false;
  options->language_ = static_cast<char*>(opt[OPT_LANGUAGE]);
  options->input_ = static_cast<char*>(opt[OPT_INPUT]);
  options->output_ = static_cast<char*>(opt[OPT_OUTPUT]);
  options->hasNamespace_ = opt[OPT_NAMESPACE] ? true : false;
  options->hasRpcPortLib_ = opt[OPT_RPCLIB] ? true : false;

  return options;
}

}  // namespace

int main(int argc, char** argv) {
  std::unique_ptr<::Options> options = ::Options::Parse(argc, argv);
  if (!options)
    exit(1);

  tidl::Parser ps;
  std::string path(options->GetInput());

  if (!ps.ParseFromFile(path))
    exit(1);

  if (options->IsProxy()) {
    if (options->GetLanguage() == "C#") {
      tidl::CsProxyGen proxy(ps.GetDoc());
      proxy.Run(options->GetOutput() + ".cs");
      if (options->HasRpcPortLib()) {
        tidl::CsLibGen lib(ps.GetDoc());
        lib.Run("rpc-port.cs");
      }
    } else if (options->GetLanguage() == "C++") {
      tidl::CppProxyHeaderGen proxy_header(ps.GetDoc());
      proxy_header.Run(options->GetOutput() + ".h");
      tidl::CppProxyBodyGen proxy_body(ps.GetDoc());
      proxy_body.Run(options->GetOutput() + ".cc");
    } else if (options->GetLanguage() == "C") {
      tidl::CProxyHeaderGen proxy_header(ps.GetDoc());
      proxy_header.EnableNamespace(options->HasNamespace());
      proxy_header.Run(options->GetOutput() + ".h");
      tidl::CProxyBodyGen proxy_body(ps.GetDoc());
      proxy_body.EnableNamespace(options->HasNamespace());
      proxy_body.Run(options->GetOutput() + ".c");
    }
  } else {
    if (options->GetLanguage() == "C#") {
      tidl::CsStubGen stub(ps.GetDoc());
      stub.Run(options->GetOutput() + ".cs");
      if (options->HasRpcPortLib()) {
        tidl::CsLibGen lib(ps.GetDoc());
        lib.Run("rpc-port.cs");
      }
    } else if (options->GetLanguage() == "C++") {
      tidl::CppStubHeaderGen stub_header(ps.GetDoc());
      stub_header.Run(options->GetOutput() + ".h");
      tidl::CppStubBodyGen stub_body(ps.GetDoc());
      stub_body.Run(options->GetOutput() + ".cc");
    } else if (options->GetLanguage() == "C") {
      tidl::CStubHeaderGen stub_header(ps.GetDoc());
      stub_header.EnableNamespace(options->HasNamespace());
      stub_header.Run(options->GetOutput() + ".h");
      tidl::CStubBodyGen stub_body(ps.GetDoc());
      stub_body.EnableNamespace(options->HasNamespace());
      stub_body.Run(options->GetOutput() + ".c");
    }
  }

  return 0;
}
