%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "idlc/parser.h"
#include "idlc/document.h"
#include "idlc/declaration.h"
#include "idlc/type.h"
#include "idlc/parameter.h"
#include "idlc/interface.h"
#include "idlc/element.h"
#include "idlc/structure.h"
#include "idlc/block.h"
#include "idlc/attribute.h"
#include "idlc/tidlc_y.hpp"

int yylex(yy::parser::semantic_type *, yy::parser::location_type *, void *);

#define lex_scanner ps->Scanner()

%}

%parse-param { tidl::Parser* ps }
%lex-param { void *lex_scanner }

%token T_LEFT T_RIGHT T_COMMA T_SEMICOLON T_BRACE_OPEN T_BRACE_CLOSE
%token T_IN T_OUT T_REF T_ASYNC
%token T_META_OPEN T_META_CLOSE
%token T_EQUAL T_DELEGATE


%start start

%skeleton "glr.cc"

%union {
  tidl::Document* doc;
  tidl::Interface* interf;
  tidl::Declaration* decl;
  tidl::Declarations* decls;
  tidl::BaseType* b_type;
  tidl::ParameterType* p_type;
  tidl::Token* direction;
  tidl::Parameter* param;
  tidl::Parameters* params;
  tidl::Token* token;
  tidl::Structure* structure;
  tidl::Element* elm;
  tidl::Elements* elms;
  tidl::Block* blk;
  tidl::Attribute* attr;
  tidl::Attributes* attrs;
}

%token<token> T_ID
%token<token> T_STRUCTURE
%token<token> T_INTERFACE
%token<token> T_CHAR
%token<token> T_SHORT
%token<token> T_INT
%token<token> T_LONG
%token<token> T_FLOAT
%token<token> T_DOUBLE
%token<token> T_VOID
%token<token> T_BUNDLE
%token<token> T_STRING
%token<token> T_BOOL
%token<token> T_LIST
%token<token> T_ARRAY
%token<token> T_VALUE
%token<token> T_SB_OPEN
%token<token> T_SB_CLOSE
%type<doc> blocks
%type<blk> block
%type<structure> structure_block
%type<interf> interface_block
%type<decls> declarations
%type<decl> declaration
%type<p_type> parameter_type
%type<b_type> base_type
%type<b_type> container_type
%type<param> parameter
%type<params> parameter_list
%type<direction> direction_specifier
%type<token> container_type_name
%type<elm> element
%type<elms> elements
%type<attr> attribute
%type<attrs> attributes

%%

start: blocks {
     ps->SetDoc($1);
  }
;

blocks: block {
    $$ = new tidl::Document();
    if ($1 != NULL)
      $$->AddBlock($1);
  }
  | blocks block {
    $$ = $1;

    if ($2 != NULL) {
      if ($$->ExistBlock($2)) {
        ps->ReportError("syntax error. \"Already Exists\".", $2->GetLine());
        delete $2;
      } else {
        $$->AddBlock($2);
      }
    }
  }
;

block: interface_block {
    $$ = $1;
  }
  | structure_block {
    $$ = $1;
  }
;

structure_block: T_STRUCTURE T_ID T_BRACE_OPEN elements T_BRACE_CLOSE {
    $$ = new tidl::Structure($2->ToString(), $4, $1->GetComments(),
        @1.begin.line);
    delete $1;
    delete $2;
  }
  | T_STRUCTURE T_BRACE_OPEN elements T_BRACE_CLOSE {
    ps->ReportError("syntax error. \"No identifier\".", @1.begin.line);
    $$ = NULL;
    delete $1;
  }
  | T_STRUCTURE error T_BRACE_OPEN elements T_BRACE_CLOSE {
    ps->ReportError("syntax error. \"Please check it before an open brace.\"",
        @2.begin.line);
    $$ = NULL;
    delete $1;
  }
  | T_STRUCTURE error T_BRACE_CLOSE {
    ps->ReportError("syntax error in structure declaration.", @2.begin.line);
    $$ = NULL;
    delete $1;
  }
;

elements: element {
    $$ = new tidl::Elements();
    $$->Add($1);
  }
  | elements element {
    $$ = $1;
    if ($$->Exist($2)) {
      ps->ReportError("syntax error. \"Already Exists\".", $2->GetLine());
      delete $2;
    } else {
      $$->Add($2);
    }
  }
  | elements error T_SEMICOLON {
    ps->ReportError("syntax error in elements declarations.", @1.begin.line);
    $$ = $1;
  }
;

element: base_type T_ID T_SEMICOLON {
    $$ = new tidl::Element($2->ToString(), $1, $1->GetComments(),
        @1.begin.line);
    delete $2;
  }
  | base_type T_SEMICOLON {
    ps->ReportError("syntax error. \"No identifier\".", @2.begin.line);
    $$ = NULL;
  }
  | base_type error T_SEMICOLON {
    ps->ReportError("syntax error in element declaration.", @2.begin.line);
    $$ = NULL;
  }
  | error {
    ps->ReportError("syntax error in element declaration.", @1.begin.line);
    $$ = NULL;
  }
;

attributes: attribute {
    $$ = new tidl::Attributes();
    $$->Add($1);
  }
  | attributes T_COMMA attribute {
    $$ = $1;
    if ($$->Exist($3)) {
      ps->ReportError("syntax error. \"Already Exist\".", $3->GetLine());
      delete $3;
    } else {
      $$->Add($3);
    }
  }
  | error {
    ps->ReportError("syntax error in attributes", @1.begin.line);
    $$ = new tidl::Attributes();
  }
;

attribute: T_ID T_EQUAL T_VALUE {
    $$ = new tidl::Attribute($1->ToString(), $3->ToString(), @1.begin.line);
    delete $1;
    delete $3;
  }
  | T_ID error T_VALUE  {
    ps->ReportError("syntax error in attribute declaration.", @2.begin.line);
    $$ = NULL;
    delete $1;
    delete $3;
  }
  | T_ID T_EQUAL error {
    ps->ReportError("syntax error in attribute declaration.", @3.begin.line);
    $$ = NULL;
    delete $1;
  }
;

interface_block: T_INTERFACE T_ID T_BRACE_OPEN declarations T_BRACE_CLOSE {
    $$ = new tidl::Interface($2->ToString(), $4, $1->GetComments(),
        new tidl::Attributes(), @1.begin.line);
    delete $1;
    delete $2;
  }
  | T_SB_OPEN attributes T_SB_CLOSE T_INTERFACE T_ID T_BRACE_OPEN declarations T_BRACE_CLOSE {
    $$ = new tidl::Interface($5->ToString(), $7, $1->GetComments(), $2,
        @1.begin.line);
    delete $1;
    delete $3;
    delete $4;
    delete $5;
  }
  | T_INTERFACE T_BRACE_OPEN declarations T_BRACE_CLOSE {
    ps->ReportError("syntax error. \"No identifier\".", @1.begin.line);
    $$ = NULL;
    delete $1;
  }
  | T_INTERFACE error T_BRACE_OPEN declarations T_BRACE_CLOSE {
    ps->ReportError("syntax error in interface declaration.", @2.begin.line);
    $$ = NULL;
    delete $1;
  }
  | T_INTERFACE error T_BRACE_CLOSE {
    ps->ReportError("syntax error in interface declaration.", @2.begin.line);
    $$ = NULL;
    delete $1;
  }
;

declarations: declaration {
    $$ = new tidl::Declarations();
    $$->Add($1);
  }
  | declarations declaration {
    $$ = $1;
    if ($$->Exist($2)) {
      ps->ReportError("syntax error. \"Already Exists\".", $2->GetLine());
      delete $2;
    } else {
      $$->Add($2);
    }
  }
  | declarations error T_SEMICOLON {
    ps->ReportError("syntax error in methods declaration.", @2.begin.line);
    $$ = $1;
  }
;

declaration: base_type T_ID T_LEFT parameter_list T_RIGHT T_SEMICOLON {
    $$ = new tidl::Declaration($2->ToString(), $1, $4, $1->GetComments(),
        @1.begin.line, tidl::Declaration::MethodType::SYNC);
    delete $2;
  }
  | T_VOID T_ID T_LEFT parameter_list T_RIGHT T_ASYNC T_SEMICOLON {
    $$ = new tidl::Declaration($2->ToString(),
        new tidl::BaseType("void", $1->GetComments()), $4,
        $1->GetComments(), @1.begin.line, tidl::Declaration::MethodType::ASYNC);
    delete $1;
    delete $2;
  }
  | T_VOID T_ID T_LEFT parameter_list T_RIGHT T_DELEGATE T_SEMICOLON {
    $$ = new tidl::Declaration($2->ToString(),
        new tidl::BaseType("void", $1->GetComments()), $4,
        $1->GetComments(), @1.begin.line,
        tidl::Declaration::MethodType::DELEGATE);
    delete $1;
    delete $2;
  }
  | base_type T_ID T_LEFT parameter_list T_RIGHT T_ASYNC T_SEMICOLON {
    ps->ReportError("syntax error in method declaration.", @2.begin.line);
    $$ = NULL;
    delete $2;
  }
  | base_type T_ID T_LEFT parameter_list T_RIGHT T_DELEGATE T_SEMICOLON {
    ps->ReportError("syntax error in method declaration.", @2.begin.line);
    $$ = NULL;
    delete $2;
  }
  | T_VOID T_ID T_LEFT parameter_list T_RIGHT T_SEMICOLON {
    ps->ReportError("syntax error. \"No async\".", @6.begin.line);
    $$ = NULL;
    delete $2;
  }
  | base_type T_LEFT parameter_list T_RIGHT T_SEMICOLON {
    ps->ReportError("syntax error. \"No identifier\".", @2.begin.line);
    $$ = NULL;
  }
  | T_VOID T_LEFT parameter_list T_RIGHT T_ASYNC T_SEMICOLON {
    ps->ReportError("syntax error. \"No identifier\".", @2.begin.line);
    $$ = NULL;
  }
  | T_VOID T_LEFT parameter_list T_RIGHT T_DELEGATE T_SEMICOLON {
    ps->ReportError("syntax error. \"No identifier\".", @2.begin.line);
    $$ = NULL;
  }
  | base_type error T_SEMICOLON {
    ps->ReportError("syntax error in method declaration.", @2.begin.line);
    $$ = NULL;
  }
  | T_VOID error T_SEMICOLON {
    ps->ReportError("syntax error in method declaration.", @2.begin.line);
    $$ = NULL;
  }
;

parameter_list: parameter {
    $$ = new tidl::Parameters();
    if ($1 != nullptr)
      $$->Add($1);
  }
  | parameter_list T_COMMA parameter {
    $$ = $1;
    if ($$->Exist($3)) {
      ps->ReportError("syntax error. \"Already Exists\".", $3->GetLine());
      delete $3;
    } else {
      $$->Add($3);
    }
  }
  | error {
    ps->ReportError("syntax error in parameter list", @1.begin.line);
    $$ = new tidl::Parameters();
  }
;

direction_specifier: T_IN {
    $$ = new tidl::Token("in", "");
  }
  | T_OUT {
    $$ = new tidl::Token("out", "");
  }
  | T_REF {
    $$ = new tidl::Token("ref", "");
  }
;

parameter: {
    $$ = nullptr;
  }
  | T_VOID {
    $$ = nullptr;
    delete $1;
  }
  | parameter_type T_ID {
    $$ = new tidl::Parameter($2->ToString(), $1, @1.begin.line);
    delete $2;
  }
;

parameter_type: base_type {
      $$ = new tidl::ParameterType($1);
    }
    | direction_specifier base_type {
      $$ = new tidl::ParameterType($2, $1->ToString());
      delete $1;
    }
;

base_type: container_type {
      $$ = $1;
    }
    | T_VOID {
      $$ = new tidl::BaseType("void", $1->GetComments());
      delete $1;
    }
    | T_CHAR {
      $$ = new tidl::BaseType("char", $1->GetComments());
      delete $1;
    }
    | T_SHORT {
      $$ = new tidl::BaseType("short", $1->GetComments());
      delete $1;
    }
    | T_INT {
      $$ = new tidl::BaseType("int", $1->GetComments());
      delete $1;
    }
    | T_LONG {
      $$ = new tidl::BaseType("long", $1->GetComments());
      delete $1;
    }
    | T_FLOAT {
      $$ = new tidl::BaseType("float", $1->GetComments());
      delete $1;
    }
    | T_DOUBLE {
      $$ = new tidl::BaseType("double", $1->GetComments());
      delete $1;
    }
    | T_BUNDLE {
      $$ = new tidl::BaseType("bundle", $1->GetComments());
      delete $1;
    }
    | T_STRING {
      $$ = new tidl::BaseType("string", $1->GetComments());
      delete $1;
    }
    | T_BOOL {
      $$ = new tidl::BaseType("bool", $1->GetComments());
      delete $1;
    }
    | T_ID {
      $$ = new tidl::BaseType($1->ToString(), $1->GetComments(), true);
      delete $1;
    }
;

container_type: container_type_name T_META_OPEN base_type T_META_CLOSE {
      $$ = new tidl::BaseType($1->ToString(), $1->GetComments());
      $$->SetMetaType($3);
      delete $1;
    }
;

container_type_name: T_LIST {
      $$ = new tidl::Token("list", $1->GetComments());
      delete $1;
    }
    | T_ARRAY {
      $$ = new tidl::Token("array", $1->GetComments());
      delete $1;
    }
;

%%

#include <ctype.h>
#include <stdio.h>

void yy::parser::error(const yy::parser::location_type& l,
                       const std::string& errstr) {
  ps->ReportError(errstr, l.begin.line);
}


